#!/usr/bin/env python3

# This script, which is not part of the pdfannots package, allows pdfannots
# to by run directly from a source tree clone.

if __name__ == '__main__':
    import pdfannots.cli
    pdfannots.cli.main()
